###########################################################################
## Filename:      start.S
## Version:       $Id: start.S,v 1.15 1999/10/03 22:03:52 erikm Exp $
## Copyright:     Copyright (C) 1999, Erik Mouw
## Author:        Erik Mouw <J.A.K.Mouw@its.tudelft.nl>
## Description:   blob start code
## Created at:    Sun Jul 18 20:29:08 1999
## Modified by:   Erik Mouw <J.A.K.Mouw@its.tudelft.nl>
## Modified at:   Sun Oct  3 21:07:36 1999
## More mods by:  Jeff Sutherland, Accelent Systems, Inc. March 2000
###########################################################################
/*
 * start.S: blob start code
 *
 * Copyright (C) 1999  Erik Mouw (J.A.K.Mouw@its.tudelft.nl) and 
 *                     Jan-Derk Bakker (J.D.Bakker@its.tudelft.nl)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
/*
 * This is the blob start code. The SA-1100 jumps to address 0x00000000
 * after a reset. We put a single branch code at this position which jumps
 * to a safe region to do the actual setup. All other vectors just point
 * to an endless loop for the moment.
 *
 * Documentation: 
 * [1] Intel Corporation, "Intel StrongARM SA-1100 Microprocessor
 *     Developer's Manual", April 1999
 * [2] S. Furber, "ARM System Architecture", Addison Wesley Longman
 *     Ltd., Essex, England, 1996.
 * [3] Intel Corporation, "Intel StrongARM SA-1110 Microprocessor
 *     Advanced Developer's manual, December 1999
 */

.ident "$Id: start.S,v 1.20 2000/03/14 08:48:00 erikm/jws Exp $"

.text

/* extern function to blink the LED */
.extern blink_led


/* Jump vector table as in table 3.1 in [1] */
.globl _start
_start:	b	reset
	b	undefined_instruction
	b	software_interrupt
	b	abort_prefetch
	b	abort_data
	b	not_used
	b	irq
	b	fiq
	
reset:

	.rept	64
	nop
	.endr
/* do that to make sure we're awake... */
old_blink_led:	
	/* Register addresses can be found in [1] Appendix A */
/* allow 32kHz clock to settle before continuing... */
/* Also, provides the 100-200uSec delay required by SDRAM clocks */

	mov	r1, #0x90000000
	add	r1, r1, #0x20000
1:	ldr	r2, [r1, #0x1c]
	tst	r2, #1
	beq	1b

	/* First, mask **ALL** interrupts */
	mov	r1, #0x90000000
	add	r1, r1, #0x50000
	mov	r2, #0x00
	str	r2, [r1, #0x4]

	mov	r3, #0x10000000		/* address of hex LEDs */
	mov	r2, #1
	str	r2, [r3, #0x10]
        add	r2, r2, #0xFE
	str	r2, [r3, #0x50]		/* don't forget to turn them on! */
	nop
	str	r2, [r3, #0x50]

	/* Switch the SA-1100/SA-1110 to 132 MHz (5) or 118 MHz (4) */
	mov	r1, #0x90000000
	add	r1, r1, #0x20000
	mov	r2, #0x5
	str	r2, [r1, #0x14]

/* now it gets tricky: */


	/* Set up the SDRAM in banks 0 and 1 */
	mov	r1, #0xA0000000		/* MDCNFG */
@	mov	r2,     #0xF4000000	/* start with sdram disabled */
@	add	r2, r2, #0x00500000
@	add	r2, r2, #0x00007200
@	add	r2, r2, #0x00000054
@	str	r2, [r1, #0x00]
					/* MDREFR */
@	mov	r2,     #0x49000000
@	add	r2, r2, #0x00BC0000
@	add	r2, r2, #0x00000300
@	add	r2, r2, #0x00000027
@	str	r2, [r1, #0x1C]
					/* MDCAS00 */
	mov	r2,     #0xAA000000
	add	r2, r2, #0x00AA0000
	add	r2, r2, #0x0000AA00
	add	r2, r2, #0x0000007F
	str	r2, [r1, #0x04]
					/* MDCAS01 */
	mov	r2,     #0xAA000000
	add	r2, r2, #0x00AA0000
	add	r2, r2, #0x0000AA00
	add	r2, r2, #0x000000AA
	str	r2, [r1, #0x08]	
					/* MDCAS02 */
	mov	r2,     #0xAA000000
	add	r2, r2, #0x00AA0000
	add	r2, r2, #0x0000AA00
	add	r2, r2, #0x000000AA
	str	r2, [r1, #0x0C]	
	b	10f

/* stolen from the ce bootloader:
; MDREFR bit fields
DRI		EQU	0xFFF0
TRASR		EQU	0x000F
;E0PIN		EQU	BIT16
;K0RUN		EQU	BIT17
K0DB2		EQU	BIT18
E1PIN		EQU	BIT20
K1RUN		EQU	BIT21
K1DB2		EQU	BIT22
K2RUN		EQU	BIT25
K2DB2		EQU	BIT26
EAPD		EQU	BIT28
KAPD		EQU	BIT29
SLFRSH		EQU	BIT31

;MDREFR_MASK	EQU	DRI:OR:TRASR:OR:K0DB2:OR:K1DB2:OR:K2DB2:OR:EAPD:OR:KAPD
MDREFR_MASK	EQU	DRI:OR:TRASR:OR:K0DB2:OR:K1DB2:OR:K2DB2:OR:EAPD:OR:KAPD:OR:K1RUN:OR:K2RUN:OR:E1PIN
*/
#define MDREFR_MASK #0xb674ffff

	@ sequence the SDRAM controller from "self-refresh and clock stop" to "idle"
	@ - self-refresh and clock stop (state after reset)
	@ - self-refresh
	@ - power-down
	@ - PWRDNX
	@ - idle

	@ SDRAM command is NOP or COMMAND_INHIBIT, SDRAM CONTROLLER: self-refresh and clock stop

	@ First get the address of and starting value for the MDREFR register
10:	mov	r4, #0xa0000000
	add	r4, r4, #0x1C
@	ldr	r4, =MDREFR_REAL
	ldr	r5, [r4]

	@ clear fields to be updated
@	ldr	r6, =MDREFR_MASK
	mov	r6, #0xb6000000
	add	r6, r6, #0x740000
	add	r6, r6, #0xff00
	add	r6, r6, #0xff
	bic	r5, r5, r6

	@ set KnDB2, DRI, TRAS fields to the correct values, keep auto power down bits disabled
	@ ??? Must use proper value based on selected CPU speed 0 or 1
@	ldr	r6, =(DRAM_MDREFR_VALUE_0:AND:(:NOT:(EAPD:OR:KAPD)))
	mov	r6, #0x04000000
	add	r6, r6, #0x470000
	add	r6, r6, #0x0100
	add	r6, r6, #0xe7
	orr	r5, r5, r6
	str	r5, [r4]

	@ SDRAM CONTROLLER: move from "self-refresh and clock stop" to "self-refresh" (set K1RUN and K2RUN bits)
@	orr	r5, r5, #(K1RUN:OR:K2RUN)
	orr	r5, r5, #0x02200000
	str	r5, [r4]
		
	@ SDRAM command is NOP or COMMAND_INHIBIT, SDRAM CONTROLLER: self-refresh

	@ Delay with clock running
	mov	r0, r3
	mov	r1, r4
	mov 	r2, r6 
	bl	tenthSecDelay				@ Trashes r3,4,6
	mov	r3, r0
	mov	r4, r1
	mov 	r6, r2

	@ SDRAM CONTROLLER: move from "self-refresh" to "power-down" (clear SLFRSH bit)
	bic	r5, r5, #0x40000000
	str	r5, [r4]

	@ SDRAM command is NOP or COMMAND_INHIBIT, SDRAM CONTROLLER: power-down

	@ Delay with clock running

	mov	r0, r3
	mov	r1, r4
	mov 	r2, r6 
	bl	tenthSecDelay				@ Trashes r3,4,6
	mov	r3, r0
	mov	r4, r1
	mov 	r6, r2

	@ SDRAM CONTROLLER: move from "power-down" to "PWRDNX" (set E1PIN bit)
	orr	r5, r5, #0x00100000
	str	r5, [r4]

	@ SDRAM command is NOP or COMMAND_INHIBIT, SDRAM CONTROLLER: PWRDNX

	@ Delay with clock running
	mov	r0, r3
	mov	r1, r4
	mov 	r2, r6 
	bl	tenthSecDelay				@ Trashes r3,4,6
	mov	r3, r0
	mov	r4, r1
	mov 	r6, r2

	@ SDRAM CONTROLLER: the move from "PWRDNX" to "IDLE" is automatic
	@ SDRAM command is COMMAND_INHIBIT (NOP)
	

	@ For SDRAM, the first refresh command will also cause a PAL (Precharge All banks)
	@  before the CBR (auto refresh)

	@ Cause CBR DRAM/SDRAM cycles to be run by non-burst read of any disabled DRAM/SDRAM bank
@	ldr	r4, =DRAM_BANK0_REAL
	mov	r4, #0xC0000000

	@ Loop for selected number of CBR (SDRAM auto refresh) refresh cycles
@	ldr	r5, =DRAM_CBR_BURST_COUNT
	mov	r7, #16		@need at least 8 as I recall...
10:
	@ Trigger CBR refresh by causing a non-burst read from a disabled DRAM/SDRAM bank
	ldr	r0, [r4]

	@ Wait for refresh to complete
@	mov	r0, #16		; about 4 uS delay

	mov	r0, r3
	mov	r1, r4
	mov 	r2, r6 
	bl	tenthSecDelay				@ Trashes r3,4,6
	mov	r3, r0
	mov	r4, r1
	mov 	r6, r2

	subs	r7, r7, #1
	bne	10b

	@ Get the system clock speed index into r0
@	bl	GetSystemSpeedIndex
	
	@ Select proper MDCNFG register value for the chosen CPU speed
@	cmp	r0, #0
@	ldreq	r1, =MDCNFG_VALUE_0
@	ldrne	r1, =MDCNFG_VALUE_1
	mov	r1, #0x7100
	add	r1, r1, #0x2f

	@ enable DRAM/SDRAM banks, for SDRAM this will also cause a MRS command which programs the SDRAM mode register
@	ldr 	r0, =MDCNFG_REAL
	mov	r0, #0xa0000000
	str	r1, [r0]

	@ Wait a bit more
	mov	r0, r3
	mov	r1, r4
	mov 	r2, r6 
	bl	tenthSecDelay				@ Trashes r3,4,6
	mov	r3, r0
	mov	r4, r1
	mov 	r6, r2

	@ Turn on auto power down bits
@	orr 	r5, r5, #(DRAM_MDREFR_VALUE_0:AND:(EAPD:OR:KAPD))
	orr	r5, r5, #0x30000000
	str	r5, [r4]

	mov	r3, #0x10000000		/* address of hex LEDs */
	mov	r2, #0x00000000
	add	r2, r2, #0x02
	str	r2, [r3, #0x10]

/* All memory settings should be ready to go... */

	/* Testing ,Chester */
	mov r3,#0x12000000
	mov r2,#0x5000                 @ D9_LED on and D8_LED off
	str r2,[r3]
	mov r4, #0x20000
gogogo2:
	subs r4, r4, #1
	bne gogogo2


	/* modified by Chester ,using serial port 1 on Assabet */
	ldr r0,=0x80020060 				@GPCLK/UART select,select UART
	mov r2,#1
	str r2,[r0]

	/* Now clear all 'sticky' bits in serial I registers, cf. [1] 11.11 */
	mov	r1, #0x80000000
	add	r1, r1, #0x10000
	mov	r2, #0xFF
	str	r2, [r1, #0x1C]
	
	/* Set the serial port to sensible defaults: no break, no interrupts, */
	/* no parity, 8 databits, 1 stopbit. */
	mov	r2, #0x00
	str	r2, [r1, #0x0C]
	mov	r2, #0x08
	str	r2, [r1, #0x00]

	/* Set BRD to 5, for a baudrate of 38k4 ([1] 11.11.4.1) */
	/* Set BRD to 23, for a baudrate of 9k6 ([1] 11.11.4.1) */
	mov	r2, #0x00
	str	r2, [r1, #0x04]
	mov	r2, #23
	str	r2, [r1, #0x08]

	/* Initialize the GPDR (GPIO Pin Direction Register) in such a
	way that the LED is on an output port */

	/* load the GPIO base in r2 */	
	mov	r2, #0x90000000
	add	r2, r2, #0x40000

	/* Bit 23 is the LED, according to JDB, but 17 on Zilker */
/*	mov	r1, #0x00800000 */
	mov	r1, #0x00020000
	str	r1, [r2, #0x04]

	/* The LED controls FORCE_ON, so turn it on here */
	str     r1, [r2, #0x08]

	/* Prepare to send some characters out in a loop */
	mov	r1, #0x80000000
	add	r1, r1, #0x10000	
	
	/* Enable the transmitter */
	mov	r2, #0x02
	str	r2, [r1, #0x0C]

	/* Send out a welcome message */
	adr	r0, welcome
	bl	print_str

	/* Determine whether this code runs from main flash or external flash */
	/* Clue: Main flash is 32 bit, external is 16 bit. */
/* don't do this for Zilker:
	mov	r1, #0xA0000000
	ldr	r0, [r1, #0x10]
	ands	r0, r0, #0x04
	bne	cpy_ext2int
*/
	/* Put the main flash to the correct speed */
	mov	r1, #0xA0000000
	mov	r2,     #0x6D000000
	add	r2, r2, #0x00680000
	add	r2, r2, #0x00006D00
	add	r2, r2, #0x00000068
	str	r2, [r1, #0x10]	/* MSC0 (for 160S3 flash chips) */

	adr	r0, int_flash_str
	bl	print_str

test_mem:
	/* Start the memory tester, hardcoded bank 0 & 1 for now */
	adr	r0, start_test
	bl	print_str

	/* Erase memory */
	adr	r0, zeroing_mem
	bl	print_str

	mov	r4, #0xC0000000
	mov	r5, #0xD0000000
	mov	r6, #0x400000
	sub	r6, r6, #0x01

	mov	r0, r4
	bl	print_hex

zero_loop:
	mov	r0, #0x0
	mov	r1, #0x0
	mov	r2, #0x0
	mov	r3, #0x0
	mov	r7, #0x0
	mov	r8, #0x0
	mov	r9, #0x0
	mov	r10, #0x0
zero_loop1:
	stmia	r4!, {r0-r3, r7-r10}
	stmia	r4!, {r0-r3, r7-r10}
	stmia	r4!, {r0-r3, r7-r10}
	stmia	r4!, {r0-r3, r7-r10}
	tst	r4, r6
	bne	zero_loop1
	adr	r0, del10
	bl	print_str
	mov	r0, r4
	bl	print_hex
	cmp	r4, r5
	blt	zero_loop

	adr	r0, zeroing_done
	bl	print_str

	mov	r4, #0xC0000000
	mov	r5, #0xD0000000
	mov	r6, r4 /* Base for current block */
	mov	r7, #0x0 /* Diff between r4 and [r4] */

	/* Go hunting for aliases */
	mov	r3, #0x10000000		/* address of hex LEDs */
	mov	r2, #0x00000000
	add	r2, r2, #0x03
	str	r2, [r3, #0x10]

alias_loop:
	ldr	r8, [r4]
	cmp	r8, #0x0
	streq	r4, [r4] /* If the value was zero, load r4 to it */
	moveq	r8, r4
	sub	r9, r4, r8
	cmp	r9, r7 /* Is this a new block ? */
	bne	alias_detected
	add	r4, r4, #0x1000
	cmp	r4, r5
	blt	alias_loop	
	b	alias_done

alias_detected:

	mov	r3, #0x10000000		/* address of hex LEDs */
	mov	r2, #0x00000000
	add	r2, r2, #0x04
	str	r2, [r3, #0x10]

	adr	r0, block_at_str
	bl	print_str
	mov	r0, r6
	bl	print_hex
	adr	r0, length_str
	bl	print_str
	sub	r0, r4, r6
	bl	print_hex
	sub	r8, r6, r7
	cmp	r7, #0x0	/* Is this a 'new' block ? */
	moveq	r10, r6		/* If so, remember the start of this block...*/ 
	moveq	r11, r4		/* ...and the start of the next block */
	mov	r6, r4
	mov	r7, r9
	beq	alias_loop
	adr	r0, alias_for_str
	bl	print_str
	mov	r0, r8
	bl	print_hex
	b	alias_loop

alias_done:

	/* Set up the stack pointer */
	sub	sp, r11, #0x04

	/* Arg 0 of the C code is the start of the block it can use; */
	/* arg 1 is the size of that block. */
	mov	r0, r10
	sub	r1, r11, r10

	/* address of hex LEDs */
@	mov	r3, #0x10000000		
@	mov	r2, #0x00000000
@	add	r2, r2, #0x05
@	str	r2, [r3, #0x10]

	/* Testing ,Chester*/
	mov r3,#0x12000000
	mov r2,#0x7800          @ D9_LED off and D8_LED off
	str r2,[r3]                                       
	mov r3, #0x50000
gogogo6:
	subs    r3, r3, #1
bne gogogo6

	/* Jump to the C code */
jump_to_c:
	bl	c_main

	/* The c code should never return ! */
        /* (but if it does, we want to know about it...) */
	adr	r0, main_prog_abort
	bl	print_str
	b	reset


blinky:
	/* The old blinker */
	mov     r2, #0x90000000
	add     r2, r2, #0x40000
	mov     r1, #0x00020000

old_led_on:
	/* turn on the LED by writing the GPSR (GPIO Pin output Set
	Register) */
	str	r1, [r2, #0x08]
	mov	r4, #0x10000
loop1:
	subs	r4, r4, #1
	bne	loop1

old_led_off:
	/* turn off the LED by writing the GPCR (GPIO Pin output Clear
	Register) */
	str	r1, [r2, #0x0c]

	mov     r4, #0x10000
loop2:
	subs    r4, r4, #1
	bne	loop2

	/* and loop forever */
	b	old_led_on


.align 4
welcome:
	.string "\n\r\n\rAccelent Systems BLOB boot loader for Linux, v0.0.\n\r"

.align 4
int_flash_str:
	.string "Running from internal Flash.\n\r"

.align 4
start_test:
	.string "Starting the memory tester...\n\r"

.align 4
zeroing_mem:
	.string "Zeroing memory..."

.align 4
zeroing_done:
	.string "\n\rZeroing done.Testing aliases\n\r"

.align 4
main_prog_abort:
	.string "\n\rMain function aborted!\n\r"

.align 4
del10:
	.string "\b\b\b\b\b\b\b\b\b\b"

.align 4
block_at_str:
	.string "\n\rBlock at "

.align 4
length_str:
	.string " with length "

.align 4
alias_for_str:
	.string " is an alias of "
.align 4

	/* Subroutine that sends a string over the serial port */
	/* The address of the string should be in r0 */
print_str:
	/* Save the return address */
	mov	r13, r14
	mov	r2, r0
prs1:
	ldrsb	r0, [r2]
	add	r2, r2, #0x01
	ands	r0, r0, #0xFF
	beq	prs2
	bl	print_byte
	b	prs1

prs2:
	/* Return */
	mov	pc, r13


	/* Subroutine to send a hex word (in r0) over the serial port */
print_hex:
	mov	r13, r14
	mov	r2, r0
	mov	r3, #0x08
	mov	r0, #0x30
	bl	print_byte
	mov	r0, #0x78
	bl	print_byte
prh1:
	and	r0, r2, #0xF0000000
	mov	r0, r0, lsr #28
	add	r0, r0, #0x30
	cmp	r0, #0x3A
	addge	r0, r0, #0x07
	bl	print_byte
	mov	r2, r2, lsl #4
	subs	r3, r3, #0x01	
	bne	prh1

	mov	pc, r13

	/* Subroutine that sends a byte over the serial port. */
	/* The byte is in r0 */
print_byte:
	/* Wait for room in the tx fifo */
	mov     r1, #0x80000000
	add     r1, r1, #0x10000
	ldr     r1, [r1, #0x1C]
	ands    r1, r1, #0x01
	beq	print_byte

	mov     r1, #0x80000000
	add     r1, r1, #0x10000
	str	r0, [r1, #0x14]
	mov	pc, r14
	

cpy_ext2int:

	/* Put the Flash to the correct speed */
	mov	r1, #0xA0000000
	mov	r2,     #0x48000000
	add	r2, r2, #0x00680000
	add	r2, r2, #0x0000AD00
	add	r2, r2, #0x0000006C
	str	r2, [r1, #0x10]

	adr	r0, ext_flash_str
	bl	print_str

	/* Get the ID of the internal flash */
	mov	r0, #0x0090
	add	r0, r0, #0x00900000
	bl	data_to_flash
	mov	r4, #0x08000000
	str	r0, [r4]
	ldr	r0, [r4]
	bl	data_from_flash
	mov	r5, r0
	ldr	r0, [r4, #0x04]
	bl	data_from_flash
	mov	r6, r0
	mov	r3, #0x0089
	add	r3, r3, #0x00890000
	cmp	r3, r5
	beq	flash_man_ok
	adr	r0, int_unknown_str
	bl	print_str
	mov	r0, r5
	bl	print_hex
	b	cpy_exit
flash_man_ok:
	adr	r0, int_intel_str
	bl	print_str
	mov	r3,     #0x88000000
	add	r3, r3, #0x00F40000
	add	r3, r3, #0x00008800
	add	r3, r3, #0x000000F4
	cmp	r3, r6
	beq	flash_dev_ok
	adr	r0, int_unknown2_str
	bl	print_str
	mov	r0, r6
	bl	print_hex
	b	cpy_exit
flash_dev_ok:
	adr	r0, int_16m_b_str
	bl	print_str

	/* Find the size of the data in the external flash */
	mov	r5, #0x020000  /* Sod it, we hardcode it to 128k */

	/* Erase blocks in the internal flash, when needed */
	mov	r4, #0x08000000
	add	r6, r4, r5

	mov	r0, #0x00500000
	add	r0, r0, #0x00000050
	bl	data_to_flash
	str	r0, [r4]

erase_loop:
	/* switch to read array mode, if we're not there already */
	mov	r0, #0x00FF0000
	add	r0, r0, #0x000000FF
	bl	data_to_flash
	str	r0, [r4]
	ldr	r0, [r4]
	cmn	r0, #0x01
	beq	no_erase
	adr	r0, erase_str
	bl	print_str
	mov	r0, r4
	bl	print_hex
	mov	r0, #0x00200000
	add	r0, r0, #0x00000020
	bl	data_to_flash
	str	r0, [r4]
	mov	r0, #0x00D00000
	add	r0, r0, #0x000000D0
	bl	data_to_flash
	str	r0, [r4]
	bl	wait_status
	mov	r3, #0x00200000
	add	r3, r3, #0x00000020
	tst	r0, r3
	beq	erase_ok
	bl	print_hex
	adr	r0, erase_err_str
	bl	print_str
	b	cpy_exit
erase_ok:
	adr	r0, erase_end_str
	bl	print_str
no_erase:
	add	r4, r4, #0x04
	cmp	r4, r6
	blt	erase_loop

	/* Write blocks to the internal flash */
	mov	r4, #0x00

	mov	r0, #0x00500000
	add	r0, r0, #0x00000050
	bl	data_to_flash
	str	r0, [r4]

	adr	r0, write_start_str
	bl	print_str

write_loop:
	ldr	r7, [r4]
	cmn	r7, #0x01
	beq	no_write
	add	r6, r4, #0x08000000
	mov	r0, #0x00400000
	add	r0, r0, #0x00000040
	bl	data_to_flash
	str	r0, [r6]
	str	r7, [r6]
	bl	wait_status
	mov	r3, #0x00100000
	add	r3, r3, #0x00000010
	tst	r0, r3
	beq	no_write
	mov	r7, r0
	adr	r0, write_err1_str
	bl	print_str
	mov	r0, r7
	bl	print_hex
	adr	r0, write_err2_str
	bl	print_str
	mov	r0, r6
	bl	print_hex
	b	cpy_exit
no_write:
	add	r4, r4, #0x04
	cmp	r4, r5
	blt	write_loop

	/* Verify blocks written to the internal flash */
	mov	r4, #0x00
	add	r6, r4, #0x08000000

	mov	r0, #0x00500000
	add	r0, r0, #0x00000050
	bl	data_to_flash
	str	r0, [r6]

	mov	r0, #0x00FF0000
	add	r0, r0, #0x000000FF
	bl	data_to_flash
	str	r0, [r6]

	adr	r0, vrfy_start_str
	bl	print_str

vrfy_loop:
	ldr	r7, [r4]
	add	r6, r4, #0x08000000
	ldr	r8, [r6]
	cmp	r7, r8
	beq	next_vrfy
	adr	r0, vrfy_err1_str
	bl	print_str
	mov	r0, r6
	bl	print_hex
	adr	r0, vrfy_err2_str
	bl	print_str
	mov	r0, r7
	bl	print_hex
	adr	r0, vrfy_err3_str
	bl	print_str
	mov	r0, r8
	bl	print_hex
	b	cpy_exit
next_vrfy:
	add	r4, r4, #0x04
	cmp	r4, r5
	blt	vrfy_loop

	adr	r0, cpy_done_str
	bl	print_str

cpy_exit:
	b	test_mem

.align 4
ext_flash_str:
	.string "Running from external flash.\n\r"

.align 4
int_intel_str:
	.string "Internal flash manufacturer: Intel.\n\r"

.align 4
int_unknown_str:
	.string "Error: Unknown internal flash manufacturer ID "

.align 4
int_16m_b_str:
	.string "Internal flash devices: 2 x 28F160F3 16Mbit bottom flash.\n\r"

.align 4
int_unknown2_str:
	.string "Error: Unknown internal flash devices ID "

.align 4
erase_str:
	.string "Erasing block at "

.align 4
erase_err_str:
	.string "  <- error !\n\r"

.align 4
erase_end_str:
	.string " done.\n\r"

.align 4
write_start_str:
	.string "Writing data...\n\r"

.align 4
write_err1_str:
	.string "Error "

.align 4
write_err2_str:
	.string " at "

.align 4
vrfy_start_str:
	.string "Verifying data...\n\r"

.align 4
vrfy_err1_str: 
	.string "Verify error at address "

.align 4
vrfy_err2_str:
	.string "; expected : "

.align 4
vrfy_err3_str:
	.string	"; found : "

.align 4
cpy_done_str:
	.string "Copying done.\n\r"

.align 4
	/* Wait for the status of *both* internal flash devices to be ready. */
	/* Returns the status in r0. */
wait_status:
	mov	r13, r14
wait_status_loop:
	mov	r0, #0x00700000
	add	r0, r0, #0x0070
	bl	data_to_flash
	mov	r1, #0x08000000
	str	r0, [r1]
	ldr	r0, [r1]
	bl	data_from_flash
	tst	r0, #0x00000080
	tstne	r0, #0x00800000
	beq	wait_status_loop
	mov	pc, r13

.globl	data_to_flash
.globl	data_from_flash
	/* Subroutine that takes data in r0 and formats it so it will be in */
	/* the correct order for the internal flash */
data_to_flash:
	mov	r1, #0x0
	
	tst	r0, #0x00000001
	orrne	r1, r1, #0x00001000
	tst	r0, #0x00000002
	orrne	r1, r1, #0x00004000
	tst	r0, #0x00000004
	orrne	r1, r1, #0x00000800
	tst	r0, #0x00000008
	orrne	r1, r1, #0x00000200
	tst	r0, #0x00000010
	orrne	r1, r1, #0x00000001
	tst	r0, #0x00000020
	orrne	r1, r1, #0x00000004
	tst	r0, #0x00000040
	orrne	r1, r1, #0x00000080
	tst	r0, #0x00000080
	orrne	r1, r1, #0x00000020

	tst	r0, #0x00000100
	orrne	r1, r1, #0x00002000
	tst	r0, #0x00000200
	orrne	r1, r1, #0x00008000
	tst	r0, #0x00000400
	orrne	r1, r1, #0x00000400
	tst	r0, #0x00000800
	orrne	r1, r1, #0x00000100
	tst	r0, #0x00001000
	orrne	r1, r1, #0x00000002
	tst	r0, #0x00002000
	orrne	r1, r1, #0x00000008
	tst	r0, #0x00004000
	orrne	r1, r1, #0x00000040
	tst	r0, #0x00008000
	orrne	r1, r1, #0x00000010

	tst	r0, #0x00010000
	orrne	r1, r1, #0x00100000
	tst	r0, #0x00020000
	orrne	r1, r1, #0x00400000
	tst	r0, #0x00040000
	orrne	r1, r1, #0x00080000
	tst	r0, #0x00080000
	orrne	r1, r1, #0x00020000
	tst	r0, #0x00100000
	orrne	r1, r1, #0x01000000
	tst	r0, #0x00200000
	orrne	r1, r1, #0x04000000
	tst	r0, #0x00400000
	orrne	r1, r1, #0x80000000
	tst	r0, #0x00800000
	orrne	r1, r1, #0x20000000

	tst	r0, #0x01000000
	orrne	r1, r1, #0x00200000
	tst	r0, #0x02000000
	orrne	r1, r1, #0x00800000
	tst	r0, #0x04000000
	orrne	r1, r1, #0x00040000
	tst	r0, #0x08000000
	orrne	r1, r1, #0x00010000
	tst	r0, #0x10000000
	orrne	r1, r1, #0x02000000
	tst	r0, #0x20000000
	orrne	r1, r1, #0x08000000
	tst	r0, #0x40000000
	orrne	r1, r1, #0x40000000
	tst	r0, #0x80000000
	orrne	r1, r1, #0x10000000

	mov	r0, r1
	mov	pc, r14

	/* Takes data received from the flash, and unshuffles it. */
data_from_flash:
	mov	r1, #0x00

	tst	r0, #0x00000001
	orrne	r1, r1, #0x00000010
	tst	r0, #0x00000002
	orrne	r1, r1, #0x00001000
	tst	r0, #0x00000004
	orrne	r1, r1, #0x00000020
	tst	r0, #0x00000008
	orrne	r1, r1, #0x00002000
	tst	r0, #0x00000010
	orrne	r1, r1, #0x00008000
	tst	r0, #0x00000020
	orrne	r1, r1, #0x00000080
	tst	r0, #0x00000040
	orrne	r1, r1, #0x00004000
	tst	r0, #0x00000080
	orrne	r1, r1, #0x00000040

	tst	r0, #0x00000100
	orrne	r1, r1, #0x00000800
	tst	r0, #0x00000200
	orrne	r1, r1, #0x00000008
	tst	r0, #0x00000400
	orrne	r1, r1, #0x00000400
	tst	r0, #0x00000800
	orrne	r1, r1, #0x00000004
	tst	r0, #0x00001000
	orrne	r1, r1, #0x00000001
	tst	r0, #0x00002000
	orrne	r1, r1, #0x00000100
	tst	r0, #0x00004000
	orrne	r1, r1, #0x00000002
	tst	r0, #0x00008000
	orrne	r1, r1, #0x00000200

	tst	r0, #0x00010000
	orrne	r1, r1, #0x08000000
	tst	r0, #0x00020000
	orrne	r1, r1, #0x00080000
	tst	r0, #0x00040000
	orrne	r1, r1, #0x04000000
	tst	r0, #0x00080000
	orrne	r1, r1, #0x00040000
	tst	r0, #0x00100000
	orrne	r1, r1, #0x00010000
	tst	r0, #0x00200000
	orrne	r1, r1, #0x01000000
	tst	r0, #0x00400000
	orrne	r1, r1, #0x00020000
	tst	r0, #0x00800000
	orrne	r1, r1, #0x02000000

	tst	r0, #0x01000000
	orrne	r1, r1, #0x00100000
	tst	r0, #0x02000000
	orrne	r1, r1, #0x10000000
	tst	r0, #0x04000000
	orrne	r1, r1, #0x00200000
	tst	r0, #0x08000000
	orrne	r1, r1, #0x20000000
	tst	r0, #0x10000000
	orrne	r1, r1, #0x80000000
	tst	r0, #0x20000000
	orrne	r1, r1, #0x00800000
	tst	r0, #0x40000000
	orrne	r1, r1, #0x40000000
	tst	r0, #0x80000000
	orrne	r1, r1, #0x00400000

	mov	r0, r1
	mov	pc, r14
	


undefined_instruction:

	mov	r12, lr			@store addr of bad instruction
	mov	r10, #0x10000000
	mov 	r11, #0xaa000000
1:	str	r11, [r10, #0x10]
	bl	oneSecDelay
	str	r12, [r10, #0x10]
	bl	oneSecDelay
	b	1b
	
software_interrupt:

	mov	r10, #0x10000000
	mov 	r11, #0xbb000000
	str	r11, [r10, #0x10]
	b	software_interrupt

abort_prefetch:

	mov 	r12, lr
	mov	r10, #0x10000000
	mov 	r11, #0xcc000000
1:	str	r11, [r10, #0x10]
	bl	oneSecDelay
	str	r12, [r10, #0x10]
	bl	oneSecDelay
	b	1b

abort_data:
	
	mov	r12, lr
	mov	r10, #0x10000000
	mov 	r11, #0xdd000000
1:	str	r11, [r10, #0x10]
	bl	oneSecDelay
	str	r12, [r10, #0x10]
	bl	oneSecDelay
	b	1b

not_used:

	mov	r10, #0x10000000
	mov 	r11, #0xee000000
	str	r11, [r10, #0x10]
	b	not_used
	
irq:

	mov	r10, #0x10000000
	mov 	r11, #0xf1000000
	str	r11, [r10, #0x10]
	b	irq

fiq:
	mov	r10, #0x10000000
	mov 	r11, #0xf2000000
	str	r11, [r10, #0x10]
	b	fiq

oneSecDelay:
	mov	r3, #0x90000000
	add	r3, r3, #0x10000
	ldr	r4, [r3, #0x4] 		@get initial value
1:	ldr	r6, [r3, #0x4]
	sub	r6, r6, r4
	cmp	r6, #0x8000
	ble	1b
	mov	pc, lr

tenthSecDelay:
	mov	r3, #0x90000000
	add	r3, r3, #0x10000
	ldr	r4, [r3, #0x4]
1:	ldr	r6, [r3, #0x4]
	sub	r6, r6, r4
	cmp	r6, #0xcc0
	ble	1b
	mov	pc, lr

